-------     =================================================================
------------   Lamport solution (1974) for mutual exclusion between N processes
-------     =================================================================
-- each requestor receives a ticket with a value larger than former delivered tickets
-- there is for each process a state for ticket acquisition
-- the process elected for entering its critical section 
-- is the one holding the smallest ticket value among the requestors 
-- a requestor acquiring a ticket causes the others to wait for him during the election
-- several requestors concurrently acquiring a ticket may have the same ticket value
-- a fixed priority value (the process unique id for example) is used to order them

package body Lamport_N is
   
	-- initialisation protocol
	Ticket  : array (Global.Process_Id) of Integer := (others => 0);
	Chosing : array (Global.Process_Id) of Boolean := (others => False); -- ticket acquisition 
	---------------------------
	 -- function Max_Ticket returns the maximum of the current values traversed
	 -- since Max_Ticket is not a critical section, it can be executed concurrently
	 -- and can deliver identical values to concurrent processes
	 
	 function Max_Ticket return Integer is
      M : Integer := 0;
	begin
      for I in Global.Process_Id loop
         if Ticket(I) > M then M := Ticket(I); end if;
      end loop;
      return M;
   end Max_Ticket; 
	---------------------------
   -- function Precedence(X,Y) fixes a total order of service for Ticket holders
   
   function Precedence (X, Y : in Global.Process_Id) return Boolean is
   begin
      return 
		(Ticket (X) < Ticket (Y))
			or 
		(Ticket (X) = Ticket (Y) and Integer(X) < Integer(Y));
   end Precedence;
	---------------------------
   
   procedure Enter(Requestor : in Global.Process_Id) is 
   begin 
		-- the requestor acquires a ticket value
		Chosing(Requestor) := True;  -- starting ticket value acquisition
		Ticket(Requestor) := Max_Ticket; 
		Chosing(Requestor) := False; -- the ticket value is firm now
		
		-- the requestor compares its ticket with other requesting processes
		for K in Global.Process_Id loop
            while Chosing(K) loop
               null; -- requestor K has not a stable ticket value 
            end loop;
			-- a requesting process has a ticket value different from zero
            while Ticket(K) /= 0 and Precedence(K, Requestor) loop
               null;
            end loop; --the requestor has to wait until K leaves its critical section
         end loop; -- this requestor has priority over all other requesting processes
		 -- it can enter now its critical section 
	end Enter;
	
	procedure Quit(Requestor : in Global.Process_Id) is 
	begin
		Ticket(Requestor) := 0; -- this process is no longer requesting 
	end Quit;

end Lamport_N;